#include "local-include/reg.h"
#include <cpu/cpu.h>
#include <cpu/ifetch.h>
#include <cpu/decode.h>

extern void ftracer(int rd, int rs1, word_t addr, word_t pc);
// ftrace
#ifdef CONFIG_FTRACE
typedef struct symrecord{
	uint64_t addr;
	uint32_t szie;
	char name[50];
} symrecord;
extern size_t record_size;
extern symrecord FuncRecord[100];

int ftrace_sp = 0;

void ftrace_wirte(char *name, word_t addr, word_t pc, int ret){
  log_write("0x%lx: ",pc);
  ftrace_indent(ftrace_sp);
  if(ret)
    log_write("ret [%s]\n",name);
  else
    log_write("call [%s@0x%lx]\n",name, addr);
}

void ftracer(int rd, int rs1, word_t addr, word_t pc){
  int is_ret = 0;
  char * name = NULL;
  char * curr_name = NULL;
  for (size_t i = 0; i < record_size; i++){
    if(FuncRecord[i].addr == addr){
      name = FuncRecord[i].name;
      break;
    }
    else if((FuncRecord[i].addr < pc) && (pc < (FuncRecord[i].addr + FuncRecord[i].szie))){
      curr_name = FuncRecord[i].name;
    }
  }
  if (name == NULL && curr_name != NULL){
    is_ret = 1;
  }
  

  if(name != NULL || curr_name != NULL){
    if (is_ret){
      if(rd == 0 && rs1 == 1){
        ftrace_sp--;
        ftrace_wirte(curr_name,addr,pc,is_ret);
      }
    }
    else{
      ftrace_wirte(name, addr, pc,is_ret);
      ftrace_sp++;
    }
  }

  return;
}
#endif

#define JARL(rd, rs1, addr, pc) if(CONFIG_FTRACE) ftracer(rd,rs1,addr,pc); 

#define JAL(rd, rs1, addr, pc) if(CONFIG_FTRACE) ftracer(rd,rs1,addr,pc);

#define R(i) gpr(i)
#define Mr vaddr_read
#define Mw vaddr_write

enum {
  TYPE_I, TYPE_U, TYPE_S, TYPE_J, TYPE_R, TYPE_B,
  TYPE_N, // none
};

#define src1R(n) do { *src1 = R(n); } while (0)
#define src2R(n) do { *src2 = R(n); } while (0)
#define destR(n) do { *dest = n; } while (0)
#define src1I(i) do { *src1 = i; } while (0)
#define src2I(i) do { *src2 = i; } while (0)
#define destI(i) do { *dest = i; } while (0)

static word_t immI(uint32_t i) { return SEXT(BITS(i, 31, 20), 12); }
static word_t immU(uint32_t i) { return SEXT(BITS(i, 31, 12), 20); }
static word_t immS(uint32_t i) { return (SEXT(BITS(i, 31, 25), 7) << 5) | BITS(i, 11, 7); }
static word_t immJ(uint32_t i) { return SEXT((BITS(i, 31, 31)<<20)|(BITS(i, 30, 21)<<1)|(BITS(i, 20, 20)<<11)|(BITS(i, 19, 12)<<12), 20); }
static word_t immB(uint32_t i) { return SEXT((BITS(i, 31, 31)<<12)|(BITS(i, 30, 25)<<5)|(BITS(i, 11, 8)<<1)|(BITS(i, 7, 7)<<11), 13); }

static void decode_operand(Decode *s, word_t *dest, word_t *src1, word_t *src2, int type) {
  uint32_t i = s->isa.inst.val;
  int rd  = BITS(i, 11, 7);
  int rs1 = BITS(i, 19, 15);
  int rs2 = BITS(i, 24, 20);
  destR(rd);
  switch (type) {
    case TYPE_B: destI(immB(i)); src1I(rs1); src2I(rs2); break;
    case TYPE_I: src1I(rs1);     src2I(immI(i)); break;
    case TYPE_U: src1I(immU(i)); break;
    case TYPE_S: destI(immS(i)); src1R(rs1); src2R(rs2); break;
    case TYPE_J: src1I(immJ(i)); break;
    case TYPE_R: src1I(rs1); src2I(rs2); break;
  }
}

static int decode_exec(Decode *s) {
  word_t dest = 0, src1 = 0, src2 = 0;
  s->dnpc = s->snpc;

#define INSTPAT_INST(s) ((s)->isa.inst.val)
#define INSTPAT_MATCH(s, name, type, ... /* body */ ) { \
  decode_operand(s, &dest, &src1, &src2, concat(TYPE_, type)); \
  __VA_ARGS__ ; \
}

  INSTPAT_START();

  INSTPAT("??????? ????? ????? 000 ????? 11000 11", beq     , B, if(R(src1) == R(src2)) s->dnpc = s->pc + dest);
  INSTPAT("??????? ????? ????? 001 ????? 11000 11", bne     , B, if(R(src1) != R(src2)) s->dnpc = s->pc + dest);
  INSTPAT("??????? ????? ????? 100 ????? 11000 11", blt     , B, if((sword_t)R(src1) < (sword_t)R(src2)) s->dnpc = s->pc + dest);
  INSTPAT("??????? ????? ????? 101 ????? 11000 11", bge     , B, if((sword_t)R(src1) >= (sword_t)R(src2)) s->dnpc = s->pc + dest);
  INSTPAT("??????? ????? ????? 110 ????? 11000 11", bltu    , B, if(R(src1) < R(src2)) s->dnpc = s->pc + dest);
  INSTPAT("??????? ????? ????? 111 ????? 11000 11", bgeu    , B, if(R(src1) >= R(src2)) s->dnpc = s->pc + dest);
  INSTPAT("??????? ????? ????? 000 ????? 00000 11", lb      , I, R(dest) = SEXT(Mr(R(src1) + SEXT(src2,12), 1), 8));
  INSTPAT("??????? ????? ????? 001 ????? 00000 11", lh      , I, R(dest) = SEXT(Mr(R(src1) + SEXT(src2,12), 2), 16));
  INSTPAT("??????? ????? ????? 010 ????? 00000 11", lw      , I, R(dest) = SEXT(Mr(R(src1) + SEXT(src2,12), 4),32));
  INSTPAT("??????? ????? ????? 011 ????? 00000 11", ld      , I, R(dest) = Mr(R(src1) + SEXT(src2,12), 8));
  INSTPAT("??????? ????? ????? 100 ????? 00000 11", lbu     , I, R(dest) = Mr(R(src1) + SEXT(src2,12), 1));
  INSTPAT("??????? ????? ????? 101 ????? 00000 11", lhu     , I, R(dest) = Mr(R(src1) + SEXT(src2,12), 2));
  INSTPAT("??????? ????? ????? 110 ????? 00000 11", lwu     , I, R(dest) = Mr(R(src1) + SEXT(src2,12), 4));

  INSTPAT("??????? ????? ????? 000 ????? 00100 11", addi    , I, R(dest) = R(src1) + src2);
  INSTPAT("000000? ????? ????? 001 ????? 00100 11", slli    , I, R(dest) = R(src1) << BITS(src2,6,0));
  INSTPAT("??????? ????? ????? 010 ????? 00100 11", slti    , I, R(dest) = ((sword_t) R(src1)) < ((sword_t) SEXT(src2,12)));
  INSTPAT("??????? ????? ????? 011 ????? 00100 11", sltiu   , I, R(dest) = R(src1) < SEXT(src2, 12));
  INSTPAT("??????? ????? ????? 100 ????? 00100 11", xori    , I, R(dest) = R(src1) ^ src2);
  INSTPAT("000000? ????? ????? 101 ????? 00100 11", srli    , I, R(dest) = R(src1) >> BITS(src2,6,0));
  INSTPAT("010000? ????? ????? 101 ????? 00100 11", srai    , I, R(dest) = (SEXT(BITS(R(src1),63,63),1) << (64 - BITS(src2,6,0))) | (R(src1) >> BITS(src2,6,0)));
  INSTPAT("??????? ????? ????? 110 ????? 00100 11", ori     , I, R(dest) = R(src1) | src2);
  INSTPAT("??????? ????? ????? 111 ????? 00100 11", andi    , I, R(dest) = R(src1) & src2);
  INSTPAT("010000? ????? ????? 101 ????? 00100 11", srai    , I, R(dest) = (SEXT(BITS(R(src1),63,63),1) << (64 - BITS(src2,5,0))) | (R(src1) >> BITS(src2,5,0)));
  INSTPAT("??????? ????? ????? 000 ????? 00110 11", addiw   , I, R(dest) = SEXT(BITS(R(src1) + SEXT(src2,12),31,0),32));
  INSTPAT("0000000 ????? ????? 001 ????? 00110 11", slliw   , I, R(dest) = SEXT(R(src1) << BITS(src2,5,0),32));
  INSTPAT("0000000 ????? ????? 101 ????? 00110 11", srliw   , I, R(dest) = SEXT(BITS(R(src1),31,0) >> BITS(src2,5,0),32));
  INSTPAT("0100000 ????? ????? 101 ????? 00110 11", sraiw   , I, R(dest) = SEXT((SEXT(BITS(R(src1),31,31),1) << (32 - BITS(src2,5,0))) | (R(src1) >> BITS(src2,5,0)),32));
  INSTPAT("??????? ????? ????? 000 ????? 11001 11", jalr    , I, s->dnpc = (R(src1) + SEXT(src2, 12))&(~1); R(dest) = s->pc + 4; JARL(dest,src1,s->dnpc,s->pc));

  INSTPAT("??????? ????? ????? ??? ????? 11011 11", jal     , J, s->dnpc = s->pc + src1; R(dest) = s->pc + 4; JAL(0,0,s->dnpc,s->pc));
  
  INSTPAT("0000000 ????? ????? 000 ????? 01100 11", add     , R, R(dest) = R(src1) + R(src2));
  INSTPAT("0100000 ????? ????? 000 ????? 01100 11", sub     , R, R(dest) = R(src1) - R(src2));
  INSTPAT("0000000 ????? ????? 001 ????? 01100 11", sll     , R, R(dest) = R(src1) << BITS(R(src2),5,0));
  INSTPAT("0000000 ????? ????? 010 ????? 01100 11", slt     , R, R(dest) = ((sword_t) R(src1)) < ((sword_t) R(src2)));
  INSTPAT("0000000 ????? ????? 011 ????? 01100 11", sltu    , R, R(dest) = R(src1) < R(src2));
  INSTPAT("0000000 ????? ????? 100 ????? 01100 11", xor     , R, R(dest) = R(src1) ^ R(src2));
  INSTPAT("0000000 ????? ????? 101 ????? 01100 11", srl     , R, R(dest) = R(src1) >> BITS(R(src2),5,0));
  INSTPAT("0100000 ????? ????? 101 ????? 01100 11", sra     , R, R(dest) = (SEXT(BITS(R(src1),63,63),1) << (64 - BITS(R(src2),5,0))) | (R(src1) >> BITS(R(src2),5,0)));
  INSTPAT("0000000 ????? ????? 110 ????? 01100 11", or      , R, R(dest) = R(src1) | R(src2));
  INSTPAT("0000000 ????? ????? 111 ????? 01100 11", and     , R, R(dest) = R(src1) & R(src2));
  
  INSTPAT("0000000 ????? ????? 000 ????? 01110 11", addw    , R, R(dest) = SEXT(BITS(R(src1) + R(src2),31,0),32));
  INSTPAT("0100000 ????? ????? 000 ????? 01110 11", subw    , R, R(dest) = SEXT(BITS(R(src1) - R(src2),31,0),32));
  INSTPAT("0000000 ????? ????? 001 ????? 01110 11", sllw    , R, R(dest) = SEXT(BITS(R(src1),31,0) << BITS(R(src2),4,0),32));
  INSTPAT("0000000 ????? ????? 101 ????? 01110 11", srlw    , R, R(dest) = SEXT(BITS(R(src1),31,0) >> BITS(R(src2),4,0),32));
  INSTPAT("0100000 ????? ????? 101 ????? 01110 11", sraw    , R, R(dest) = (SEXT(BITS(R(src1),31,31),1) << (32 - BITS(R(src2),4,0))) | (BITS(R(src1),31,0) >> BITS(R(src2),4,0)));

  // M Ext
  INSTPAT("0000001 ????? ????? 000 ????? 01100 11", mul     , R, R(dest) = (sword_t)R(src1) * (sword_t)R(src2));
  INSTPAT("0000001 ????? ????? 100 ????? 01100 11", div     , R, R(dest) = (sword_t)R(src1) / (sword_t)R(src2));
  INSTPAT("0000001 ????? ????? 101 ????? 01100 11", divu    , R, R(dest) = R(src1) / R(src2));
  INSTPAT("0000001 ????? ????? 110 ????? 01100 11", rem     , R, R(dest) = (sword_t)R(src1) % (sword_t)R(src2));
  INSTPAT("0000001 ????? ????? 000 ????? 01110 11", mulw    , R, R(dest) = SEXT((sword_t)SEXT(BITS(R(src1),31,0),32) * (sword_t)SEXT(BITS(R(src2),31,0),32),32));
  INSTPAT("0000001 ????? ????? 100 ????? 01110 11", divw    , R, R(dest) = SEXT((sword_t)SEXT(BITS(R(src1),31,0),32) / (sword_t)SEXT(BITS(R(src2),31,0),32),32));
  INSTPAT("0000001 ????? ????? 101 ????? 01110 11", divuw   , R, R(dest) = SEXT(SEXT(BITS(R(src1),31,0),32) / SEXT(BITS(R(src2),31,0),32),32));
  INSTPAT("0000001 ????? ????? 110 ????? 01110 11", remw    , R, R(dest) = SEXT((sword_t)SEXT(BITS(R(src1),31,0),32) % (sword_t)SEXT(BITS(R(src2),31,0),32),32));
  INSTPAT("0000001 ????? ????? 111 ????? 01110 11", remuw   , R, R(dest) = SEXT(SEXT(BITS(R(src1),31,0),32) % SEXT(BITS(R(src2),31,0),32),32));
  //

  INSTPAT("??????? ????? ????? ??? ????? 00101 11", auipc   , U, R(dest) = (src1 << 12) + s->pc);
  INSTPAT("??????? ????? ????? ??? ????? 01101 11", lui     , U, R(dest) = SEXT(src1<<12,32));

  INSTPAT("??????? ????? ????? 000 ????? 01000 11", sb      , S, Mw(src1 + dest, 1, src2));
  INSTPAT("??????? ????? ????? 001 ????? 01000 11", sh      , S, Mw(src1 + dest, 2, src2));
  INSTPAT("??????? ????? ????? 010 ????? 01000 11", sw      , S, Mw(src1 + dest, 4, src2));
  INSTPAT("??????? ????? ????? 011 ????? 01000 11", sd      , S, Mw(src1 + dest, 8, src2));
  
  INSTPAT("0000000 00001 00000 000 00000 11100 11", ebreak  , N, NEMUTRAP(s->pc, R(10))); // R(10) is $a0
  INSTPAT("??????? ????? ????? ??? ????? ????? ??", inv     , N, INV(s->pc));

  INSTPAT_END();

  R(0) = 0; // reset $zero to 0

  return 0;
}

int isa_exec_once(Decode *s) {
  s->isa.inst.val = inst_fetch(&s->snpc, 4);
  return decode_exec(s);
}
